@@ -2,13 +2,14 @@ package FFI::TinyCC;
 
 use strict;
 use warnings;
-use 5.010;
-use FFI::Raw;
+use 5.008001;
+use FFI::Platypus;
+use FFI::Platypus::Memory qw( malloc free );
 use Carp qw( croak carp );
 use File::ShareDir ();
 
 # ABSTRACT: Tiny C Compiler for FFI
-our $VERSION = '0.10'; # VERSION
+our $VERSION = '0.14'; # VERSION
 
 
 sub _dlext
@@ -18,8 +19,9 @@ sub _dlext
   $^O eq 'MSWin32' ? 'dll' : $Config::Config{dlext};
 }
 
-use constant {
-  _lib => $ENV{FFI_TINYCC_LIBTCC_SO} // (eval { File::ShareDir::dist_dir('FFI-TinyCC') } ? File::ShareDir::dist_file('FFI-TinyCC', "libtcc." . _dlext) : do {
+our $ffi = FFI::Platypus->new;
+$ffi->lib(
+  $ENV{FFI_TINYCC_LIBTCC_SO} || (eval { File::ShareDir::dist_dir('FFI-TinyCC') } ? File::ShareDir::dist_file('FFI-TinyCC', "libtcc." . _dlext) : do {
     require Path::Class::File;
     Path::Class::File
       ->new($INC{'FFI/TinyCC.pm'})
@@ -28,173 +30,81 @@ use constant {
       ->parent
       ->file('share', 'libtcc.' . _dlext)
       ->stringify
-  }),
-  
-  # tcc_set_output_type
-  _TCC_OUTPUT_MEMORY     => 0,
-  _TCC_OUTPUT_EXE        => 1,
-  _TCC_OUTPUT_DLL        => 2,
-  _TCC_OUTPUT_OBJ        => 3,
-  _TCC_OUTPUT_PREPROCESS => 4,
+  }
+));
 
-  # tcc_relocate
-  _TCC_RELOCATE_AUTO     => 1,
+$ffi->custom_type( tcc_t => {
+  perl_to_native => sub {
+    $_[0]->{handle},
+  },
   
-  # ??
-  _TCC_OUTPUT_FORMAT_ELF    => 0,
-  _TCC_OUTPUT_FORMAT_BINARY => 1,
-  _TCC_OUTPUT_FORMAT_COFF   => 2,
+  native_to_perl => sub {
+    {
+      handle   => $_[0],
+      relocate => 0,
+      error    => [],
+    };
+  },
+
+});
+
+do {
+  my %output_type = qw(
+    memory 0
+    exe    1
+    dll    2
+    obj    3
+  );
+
+  $ffi->custom_type( output_t => {
+    native_type => 'int',
+    perl_to_native => sub { $output_type{$_[0]} },
+  });
 };
 
-use constant _new => FFI::Raw->new(
-  _lib, 'tcc_new',
-  FFI::Raw::ptr,
-);
+$ffi->type('int' => 'error_t');
+$ffi->type('(opaque,string)->void' => 'error_handler_t');
 
-use constant _delete => FFI::Raw->new(
-  _lib, 'tcc_delete',
-  FFI::Raw::void,
-  FFI::Raw::ptr,
-);
+$ffi->attach([tcc_new             => '_new']             => []                                     => 'tcc_t');
+$ffi->attach([tcc_delete          => '_delete']          => ['tcc_t']                              => 'void');
+$ffi->attach([tcc_set_error_func  => '_set_error_func']  => ['tcc_t', 'opaque', 'error_handler_t'] => 'void');
+$ffi->attach([tcc_add_symbol      => '_add_symbol']      => ['tcc_t', 'string', 'opaque']          => 'int');
+$ffi->attach([tcc_get_symbol      => '_get_symbol']      => ['tcc_t', 'string']                    => 'opaque');
+$ffi->attach([tcc_relocate        => '_relocate']        => ['tcc_t', 'opaque']                    => 'int');
+$ffi->attach([tcc_run             => '_run']             => ['tcc_t', 'int', 'opaque']             => 'int');
 
-use constant _set_error_func => FFI::Raw->new(
-  _lib, 'tcc_set_error_func',
-  FFI::Raw::void,
-  FFI::Raw::ptr, FFI::Raw::ptr, FFI::Raw::ptr,
-);
-
-use constant _set_options => FFI::Raw->new(
-  _lib, 'tcc_set_options',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _add_include_path => FFI::Raw->new(
-  _lib, 'tcc_add_include_path',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _add_sysinclude_path => FFI::Raw->new(
-  _lib, 'tcc_add_sysinclude_path',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _define_symbol => FFI::Raw->new(
-  _lib, 'tcc_define_symbol',
-  FFI::Raw::void,
-  FFI::Raw::ptr, FFI::Raw::str, FFI::Raw::str,
-);
-
-use constant _undefine_symbol => FFI::Raw->new(
-  _lib, 'tcc_undefine_symbol',
-  FFI::Raw::void,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _add_file => FFI::Raw->new(
-  _lib, 'tcc_add_file',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _compile_string => FFI::Raw->new(
-  _lib, 'tcc_compile_string',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _set_output_type => FFI::Raw->new(
-  _lib, 'tcc_set_output_type',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::int,
-);
-
-use constant _add_library_path => FFI::Raw->new(
-  _lib, 'tcc_add_library_path',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _add_library => FFI::Raw->new(
-  _lib, 'tcc_add_library',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _add_symbol => FFI::Raw->new(
-  _lib, 'tcc_add_symbol',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str, FFI::Raw::ptr,
-);
-
-use constant _output_file => FFI::Raw->new(
-  _lib, 'tcc_output_file',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _run => FFI::Raw->new(
-  _lib, 'tcc_run',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::int, FFI::Raw::ptr,
-);
-
-use constant _relocate => FFI::Raw->new(
-  _lib, 'tcc_relocate',
-  FFI::Raw::int,
-  FFI::Raw::ptr, FFI::Raw::ptr,
-);
-
-use constant _get_symbol => FFI::Raw->new(
-  _lib, 'tcc_get_symbol',
-  FFI::Raw::ptr,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _set_lib_path => FFI::Raw->new(
-  _lib, 'tcc_set_lib_path',
-  FFI::Raw::void,
-  FFI::Raw::ptr, FFI::Raw::str,
-);
-
-use constant _malloc => FFI::Raw->new(
-  undef, 'malloc',
-  FFI::Raw::ptr,
-  FFI::Raw::int,
-);
-
-use constant _free => FFI::Raw->new(
-  undef, 'free',
-  FFI::Raw::void,
-  FFI::Raw::ptr,
-);
+sub _method ($;@)
+{
+  my($name, @args) = @_;
+  $ffi->attach(["tcc_$name" => "_$name"] => ['tcc_t', @args] => 'error_t');
+  eval  '# line '. __LINE__ . ' "' . __FILE__ . qq("\n) .qq{
+    sub $name
+    {
+      my \$r = _$name (\@_);
+      die FFI::TinyCC::Exception->new(\$_[0]) if \$r == -1;
+      \$_[0];
+    }
+  };
+  die $@ if $@;
+}
 
 
 sub new
 {
   my($class, %opt) = @_;
+
+  my $self = bless _new(), $class;
   
-  my $self = bless {
-    handle   => _new->call,
-    relocate => 0,
-    error    => [],
-  }, $class;
-  
-  $self->{error_cb} = FFI::Raw::Callback->new(
-    sub { push @{ $self->{error} }, $_[1] },
-    FFI::Raw::void,
-    FFI::Raw::ptr, FFI::Raw::str,
-  );
-  
-  _set_error_func->call($self->{handle}, undef, $self->{error_cb});
+  $self->{error_cb} = $ffi->closure(sub {
+    push @{ $self->{error} }, $_[1];
+  });
+  _set_error_func($self, undef, $self->{error_cb});
   
   if($^O eq 'MSWin32')
   {
     require File::Basename;
     require File::Spec;
-    my $path = File::Spec->catdir(File::Basename::dirname(_lib), 'lib');
+    my $path = File::Spec->catdir(File::Basename::dirname($ffi->lib), 'lib');
     $self->add_library_path($path);
   }
   
@@ -210,123 +120,85 @@ sub _error
   $self;
 }
 
-sub DESTROY
+if(defined ${^GLOBAL_PHASE})
 {
-  my($self) = @_;
-
-  # weird things happen during global distruction.  The
-  # _delete and _free constants go bye bye sometimes.
-  # since the process is going to end anyway, freeing
-  # the resources for this instance can be skipped.
-  if(ref(_delete) eq 'FFI::Raw' && ref(_free) eq 'FFI::Raw')
-  {  
-    _delete->call($self->{handle});
+  *DESTROY = sub
+  {
+    my($self) = @_;
+    return if ${^GLOBAL_PHASE} eq 'DESTRUCT';
+    _delete($self);
     # TODO: should we do this?
-    _free->call($self->{store}) if defined $self->{store} && !$self->{no_free_store};
+    free($self->{store});
+  }
+}
+else
+{
+  require Devel::GlobalDestruction;
+  *DESTROY = sub
+  {
+    my($self) = @_;
+    return if Devel::GlobalDestruction::in_global_destruction();
+    _delete($self);
+    # TODO: should we do this?
+    free($self->{store});
   }
 }
 
 
-sub set_options
-{
-  my($self, $options) = @_;
-  my $r = _set_options->call($self->{handle}, $options);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;
-}
+_method set_options => qw( string );
 
 
-sub add_file
-{
-  my($self, $filename) = @_;
-  my $r = _add_file->call($self->{handle}, $filename);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;
-}
+_method add_file => qw( string );
 
 
-sub compile_string
-{
-  my($self, $code) = @_;
-  my $r = _compile_string->call($self->{handle}, $code);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;
-}
+_method compile_string => qw( string );
 
 
 sub add_symbol
 {
   my($self, $name, $ptr) = @_;
-  my $r = _add_symbol->call($self->{handle}, $name, $ptr);
+  my $r;
+  if(ref($ptr) && eval { $ptr->isa('FFI::Raw::Callback') })
+  {
+    require FFI::Raw;
+    my($lib) = $ffi->lib;
+    my $add_symbol = FFI::Raw->new($lib, 'tcc_add_symbol',
+      FFI::Raw::int(),
+      FFI::Raw::ptr(), FFI::Raw::str(), FFI::Raw::ptr(),
+    );
+    $r = $add_symbol->call($self->{handle}, $name, $ptr);
+  }
+  else
+  {
+    $r = _add_symbol($self, $name, $ptr);
+  }
   die FFI::TinyCC::Exception->new($self) if $r == -1;
   $self;
 }
 
 
-sub add_include_path
-{
-  my($self, $path) = @_;
-  _add_include_path->call($self->{handle}, $path);
-  $self;
-}
+_method add_include_path => qw( string );
 
 
-sub add_sysinclude_path
-{
-  my($self, $path) = @_;
-  _add_sysinclude_path->call($self->{handle}, $path);
-  $self;
-}
+_method add_sysinclude_path => qw( string );
 
 
-sub define_symbol
-{
-  my($self, $name, $value) = @_;
-  _define_symbol->call($self->{handle}, $name, $value);
-  $self;
-}
+_method set_lib_path => qw( string );
 
 
-sub undefine_symbol
-{
-  my($self, $name) = @_;
-  _undefine_symbol->call($self->{handle}, $name);
-  $self;
-}
+$ffi->attach([tcc_define_symbol=>'define_symbol'] => ['tcc_t', 'string', 'string'] => 'void');
 
 
-my %output_type = (
-  memory => 0,
-  exe    => 1,
-  dll    => 2,
-  obj    => 3,
-);
-
-sub set_output_type
-{
-  my($self, $type) = @_;
-  croak "unknown type: $type" unless defined $output_type{$type};
-  _set_output_type->call($self->{handle}, $output_type{$type});
-  $self;
-}
+$ffi->attach([tcc_undefine_symbol=>'undefine_symbol'] => ['tcc_t', 'string', 'string'] => 'void');
 
 
-sub add_library
-{
-  my($self, $libname) = @_;
-  my $r = _add_library->call($self->{handle}, $libname);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;
-}
+_method set_output_type => qw( output_t );
 
 
-sub add_library_path
-{
-  my($self, $pathname) = @_;
-  my $r = _add_library_path->call($self->{handle}, $pathname);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;  
-}
+_method add_library => qw( string );
+
+
+_method add_library_path => qw( string );
 
 
 sub run
@@ -340,7 +212,7 @@ sub run
   my $ptrs = pack 'P' x $argc, @c_strings;
   my $argv = unpack('L!', pack('P', $ptrs));
 
-  my $r = _run->call($self->{handle}, $argc, $argv);
+  my $r = _run($self, $argc, $argv);
   die FFI::TinyCC::Exception->new($self) if $r == -1;
   $r;  
 }
@@ -352,16 +224,19 @@ sub get_symbol
   
   unless($self->{relocate})
   {
-    my $size = _relocate->call($self->{handle}, undef);
-    $self->{store} = _malloc->call($size);
-    my $r = _relocate->call($self->{handle}, $self->{store});
+    my $size = _relocate($self, undef);
+    $self->{store} = malloc($size);
+    my $r = _relocate($self, $self->{store});
     FFI::TinyCC::Exception->new($self) if $r == -1;
     $self->{relocate} = 1;
   }
-  _get_symbol->call($self->{handle}, $symbol_name);
+  _get_symbol($self, $symbol_name);
 }
 
 
+_method output_file => qw( string );
+
+
 # this variable will too be removed once this module
 # is ported to FFI::Platypus, so do not depeend on it!
 our $_get_ffi_raw_deprecation = 1;
@@ -373,18 +248,10 @@ sub get_ffi_raw
   croak "you must at least specify a return type" unless @types > 0;
   my $ptr = $self->get_symbol($symbol);
   croak "$symbol not found" unless $ptr;
+  require FFI::Raw;
   FFI::Raw->new_from_ptr($self->get_symbol($symbol), @types);
 }
 
-
-sub output_file
-{
-  my($self, $filename) = @_;
-  my $r = _output_file->call($self->{handle}, $filename);
-  die FFI::TinyCC::Exception->new($self) if $r == -1;
-  $self;
-}
-
 package
   FFI::TinyCC::Exception;
 
@@ -454,12 +321,12 @@ FFI::TinyCC - Tiny C Compiler for FFI
 
 =head1 VERSION
 
-version 0.10
+version 0.14
 
 =head1 SYNOPSIS
 
  use FFI::TinyCC;
- use FFI::Raw;
+ use FFI::Platypus::Declare qw( int );
  
  my $tcc = FFI::TinyCC->new;
  
@@ -471,14 +338,10 @@ version 0.10
    }
  });
  
- my $find_square = FFI::Raw->new_from_ptr(
-   $tcc->get_symbol('find_square'),
-   FFI::Raw::int,  # return type
-   FFI::Raw::int,  # argument types
- );
+ my $address = $tcc->get_symbol('find_square');
+ function [$address => 'find_square'] => [int] => int;
  
- # $find_square isa FFI::Raw
- say $find_square->call(4); # says 16
+ print find_square(4), "\n"; # prints 16
 
 =head1 DESCRIPTION
 
@@ -536,6 +399,18 @@ Compile a string containing C source code.
 Add the given given symbol name / callback or pointer combination.
 See example below for how to use this to call Perl from Tiny C code.
 
+It will accept a L<FFI::Raw::Callback> at a performance penalty.
+If possible pass in the pointer to the C entry point instead.
+
+If you are using L<FFI::Platypus> you can use L<FFI::Platypus#cast>
+or L<FFI::Platypus::Declare#cast> to get a pointer to a closure:
+
+ use FFI::Platypus::Declare;
+ my $clousre = closure { return $_[0]+1 };
+ my $pointer = cast '(int)->int' => 'opaque', $closure;
+ 
+ $tcc->add_symbol('foo' => $pointer);
+
 =head2 Preprocessor options
 
 =head3 add_include_path
@@ -550,6 +425,12 @@ Add the given path to the list of paths used to search for include files.
 
 Add the given path to the list of paths used to search for system include files.
 
+=head3 set_lib_path
+
+ $tcc->set_lib_path($path);
+
+Set the lib path
+
 =head3 define_symbol
 
  $tcc->define_symbol($name => $value);
@@ -601,8 +482,18 @@ Add the given directory to the search path used to find libraries.
 
  my $pointer = $tcc->get_symbol($symbol_name);
 
-Return symbol value or undef if not found.  This can be passed into
-L<FFI::Raw> or similar for use in your script.
+Return symbol address or undef if not found.  This can be passed into
+the L<FFI::Platypus#function> method, L<FFI::Platypus#attach> method,
+L<FFI::Platypus::Declare#function> function or similar interface that
+takes a pointer to a C function.
+
+=head3 output_file
+
+ $tcc->output_file($filename);
+
+Output the generated code (either executable, object or DLL) to the given filename.
+The type of output is specified by the L<set_output_type|FFI::TinyCC#set_output_type>
+method.
 
 =head3 get_ffi_raw
 
@@ -623,27 +514,13 @@ Do this:
  use FFI::Raw;
  my $function = FFI::Raw->new_from_ptr($ffi->get_symbol($name), $ret, @args);
 
-Also, L<FFI::Raw> will be replaced as a prerequisite in an upcoming version
-of L<FFI::TinyCC>, so make sure that you C<use FFI::Raw> if you are using
-it.
-
-=head3 output_file
-
- $tcc->output_file($filename);
-
-Output the generated code (either executable, object or DLL) to the given filename.
-The type of output is specified by the L<set_output_type|FFI::TinyCC#set_output_type>
-method.
+Or better yet, use L<FFI::Platypus> instead.
 
 =head1 EXAMPLES
 
 =head2 Calling Tiny C code from Perl
 
- use strict;
- use warnings;
- use 5.010;
  use FFI::TinyCC;
- use FFI::Raw;
  
  my $tcc = FFI::TinyCC->new;
  
@@ -661,48 +538,37 @@ method.
 
 =head2 Calling Perl from Tiny C code
 
- use strict;
- use warnings;
- use 5.010;
  use FFI::TinyCC;
- use FFI::Raw;
+ use FFI::Platypus::Declare qw( opaque );
  
- my $say = FFI::Raw::Callback->new(
-   sub { say $_[0] },
-   FFI::Raw::void,
-   FFI::Raw::str,
- );
+ my $say = closure { print $_[0], "\n" };
+ my $ptr = cast '(string)->void' => opaque => $say;
  
  my $tcc = FFI::TinyCC->new;
+ $tcc->add_symbol(say => $ptr);
  
- $tcc->add_symbol(say => $say);
- 
- $tcc->compile_string(q{
+ $tcc->compile_string(<<EOF);
  extern void say(const char *);
  
  int
  main(int argc, char *argv[])
  {
    int i;
-   for(i=1; i<argc; i++)
+   for(i=0; i<argc; i++)
    {
      say(argv[i]);
    }
  }
- });
+ EOF
  
- # use '-' for the program name
- my $r = $tcc->run('-', @ARGV);
+ my $r = $tcc->run($0, @ARGV);
  
  exit $r;
 
-=head2 Creating a FFI::Raw handle from a Tiny C function
+=head2 Attaching as a FFI::Platypus function from a Tiny C function
 
- use strict;
- use warnings;
- use 5.010;
  use FFI::TinyCC;
- use FFI::Raw;
+ use FFI::Platypus::Declare qw( int );
  
  my $tcc = FFI::TinyCC->new;
  
@@ -714,16 +580,14 @@ method.
    }
  });
  
- my $value = (shift @ARGV) // 4;
+ my $value = shift @ARGV;
+ $value = 4 unless defined $value;
  
- # $square isa FFI::Raw
- my $square = $tcc->get_ffi_raw(
-   'calculate_square',
-   FFI::Raw::int,  # return type
-   FFI::Raw::int,  # argument types
- );
+ my $address = $tcc->get_symbol('calculate_square');
  
- say $square->call($value);
+ function [$address => 'square'] => [int] => int;
+ 
+ print square($value), "\n";
 
 =head1 CAVEATS
 
@@ -743,10 +607,12 @@ We use the fork that comes with L<Alien::TinyCC>.
 
 =item L<Tiny C Compiler Reference Documentation|http://bellard.org/tcc/tcc-doc.html>
 
-=item L<FFI::Raw>
+=item L<FFI::Platypus>
 
 =item L<Alien::TinyCC>
 
+=item L<C::TinyCompiler>
+
 =back
 
 =head1 BUNDLED SOFTWARE
